import itertools
import numpy as np

origin = np.array([0, 0])
e1 = np.array([1, 0])
w1 = np.array([1 / 2, 3 ** .5 / 2])
w2 = np.array([5 / 6, 11 ** .5 / 6])

def count_edges(
    vertices: np.ndarray,
    atol=1e-8,
    rtol=1e-5
) -> int:
    """
    `vertices` is assumed to be a 2-dimensional float array
    representing a finite subset of a Euclidean space.
    The function returns the number of edges
    of the UDG determined by this subset.

    The optional arguments `atol` and `rtol`
    determine the numerical precision:
    given two floating point numbers `a` and `b`,
    we treat them as equal if and only if we have
    ```
    absolute(a - b) <= (atol + rtol * absolute(b)).
    ```
    """
    pass


def test_count_edges():
    for vertices, num_edges in zip(
        [
            np.array([[0]]),
            np.array([[0], [1]]),
            np.array([[0], [2]]),
            np.array([[1], [2]]),
            np.array([[1], [3]]),
            np.array([[0, 0]]),
            np.array([[0, 0], [0, 1]]),
            np.array([[0, 0], [1, 1]]),
            np.array([[0, 0], [0, 1], [0, 2]]),
            np.array([[0, 0], [1, 0], [0, 1]]),
            np.array([[2, 2], [2, 3], [5, 5], [6, 5]]),
            np.stack([origin, e1, w1]),
            np.stack([origin, e1, w1, e1 + w1, w2, w1 * w2, w2 + w1 * w2]),
            np.array([[3, 4, 5, 6], [4, 4, 5, 6], [0, 9, 9, 9], [3, 4, 5, 5]]),
            np.array(list(itertools.product(range(2), repeat=10)))
        ],
        [0, 1, 0, 1, 0, 0, 1, 0, 2, 2, 2, 3, 7, 2, 5120]
    ):
        counted_edges = count_edges(vertices)
        if not counted_edges == num_edges:
            raise ValueError(
                f"count_edges(\n\tvertices={list(map(list, vertices))}\n) is {counted_edges}, but is should be {num_edges}"
            )


    for vertices, num_edges, (atol, rtol) in zip(
        itertools.repeat(
            np.concatenate(
                [
                    np.array([[0, 0, 0], [0, 1, 0]]) + epsilon
                    for epsilon in 10. ** np.arange(-2, 2)
                ]
            ),
            400
        ),
        [4, 4, 4, 4, 6, 4, 4, 4, 4, 6, 4, 4, 4, 4, 6, 4, 4, 4, 4, 6, 6, 6, 6, 6, 6],
        itertools.product(
            10. ** np.arange(-5, 0),
            repeat=2
        )
    ):
        counted_edges = count_edges(vertices, atol=atol, rtol=rtol)
        if not counted_edges == num_edges:
            raise ValueError(
                f"count_edges(\n\tvertices={list(map(list, vertices))}\n\tatol={atol}\n\trtol={rtol}\n) is {counted_edges}, but is should be {num_edges}"
            )
        
    print("All tests passed. Congratulations!")


if __name__ == "__main__":
    test_count_edges()